"""
This module provides the asyncio client for Cisco NX-API, expected to work on:
    * N3K system
    * N5K system  (tested)
    * N7K system
    * N9K system
"""

# -----------------------------------------------------------------------------
# System Imports
# -----------------------------------------------------------------------------

from typing import Optional, List, AnyStr
import asyncio
import json
from socket import getservbyname
import ssl

# -----------------------------------------------------------------------------
# Public Imports
# -----------------------------------------------------------------------------

import httpx

# -----------------------------------------------------------------------------
# Private Imports
# -----------------------------------------------------------------------------

from asyncnxapi import xmlhelp


__all__ = ["Device"]

_ssl_context = ssl.create_default_context()
_ssl_context.options &= ~ssl.OP_NO_TLSv1_2 & ~ssl.OP_NO_TLSv1_1
_ssl_context.minimum_version = ssl.TLSVersion.TLSv1
_ssl_context.check_hostname = False
_ssl_context.verify_mode = ssl.CERT_NONE
_ssl_context.set_ciphers("HIGH:!DH:!aNULL")

# _ssl_context = ssl.SSLContext(ssl_version=ssl.PROTOCOL_TLSv1_1)  # noqa

# _ssl_context = ssl.create_default_context()
# # Sets up old and insecure TLSv1.
# _ssl_context.options &= ~ssl.OP_NO_TLSv1_3 & ~ssl.OP_NO_TLSv1_2 & ~ssl.OP_NO_TLSv1_1
# _ssl_context.minimum_version = ssl.TLSVersion.TLSv1


_NXAPI_CMD_TEMPLATE = """\
<?xml version="1.0"?>
<ins_api>
<version>{api_ver}</version>
<type>{cmd_type}</type>
<chunk>{chunk}</chunk>
<sid>{sid}</sid>
<input>{cmd_input}</input>
<output_format>{ofmt}</output_format>
</ins_api>"""


class Device(httpx.AsyncClient):
    auth: httpx.Auth = None

    DEFAULT_TIMEOUT = 60
    CMDTYPES_OPTIONS = ("cli_show", "cli_show_ascii", "cli_conf", "bash")
    OFMT_OPTIONS = ("xml", "json", "text")
    API_VER = "1.0"

    def __init__(
        self,
        host: Optional[str] = None,
        proto: Optional[str] = "https",
        port: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ):
        self.port = port or getservbyname(proto)
        self.host = host

        kwargs.setdefault("timeout", self.DEFAULT_TIMEOUT)

        if username and password:
            self.auth = httpx.BasicAuth(username, password)

        kwargs.setdefault("auth", self.auth)

        if not kwargs["auth"]:
            raise ValueError("Missing required authentication")

        kwargs.setdefault("base_url", f"{proto}://{host}:{self.port}")
        kwargs.setdefault("verify", _ssl_context)

        super(Device, self).__init__(**kwargs)

        self.headers["Content-Type"] = "application/xml"
        self.cmd_type = "cli_show"
        self.ofmt = "xml"

    def nxapi_command(self, cmd_input, formatting, sid=None):
        cmd_type = formatting.setdefault("cmd_type", self.cmd_type)
        ofmt = formatting.setdefault("ofmt", self.ofmt)

        return _NXAPI_CMD_TEMPLATE.format(
            api_ver=self.API_VER,
            cmd_type=cmd_type or self.cmd_type,
            cmd_input=cmd_input,
            chunk=0,
            sid=sid or "sid",
            ofmt=ofmt or self.ofmt,
        )

    async def nxapi_exec(self, xcmd, formatting, strip_ns=False):
        res = await self.post("/ins", data=xcmd)
        res.raise_for_status()

        if formatting["ofmt"] == "json":
            as_json = json.loads(res.text)

            outputs = as_json["ins_api"]["outputs"]["output"]
            if not isinstance(outputs, list):
                outputs = [outputs]

            # if the command results are empty then there is no 'body' in the
            # response dictionary.  in this case a return value of None
            # indicates this condition to the Caller.

            return [cmd_res.get("body") for cmd_res in outputs]

        # Output format is "xml" or "text"; but in either case the body content
        # is extracted in the same manner.

        as_text = xmlhelp.strip_ns(res.text) if strip_ns else res.text
        as_xml = xmlhelp.fromstring(as_text)

        if cli_errors := as_xml.findall(".//clierror"):
            raise RuntimeError(cli_errors)

        def body_is_text(_res_e):
            return _res_e.find("body").text.strip()

        def body_is_xml(_res):
            return _res.find("body")

        get_output = (
            body_is_text if formatting["cmd_type"] == "cli_show_ascii" else body_is_xml
        )

        return [get_output(cmd_res) for cmd_res in as_xml.xpath("outputs/output")]

    async def cli(
        self,
        command: Optional[AnyStr] = None,
        commands: Optional[List[AnyStr]] = None,
        ofmt: Optional[AnyStr] = None,
        strip_ns: Optional[bool] = False,
    ):
        if not any((command, commands)):
            raise RuntimeError("Required 'command' or 'commands'")

        formatting = dict(ofmt=ofmt or self.ofmt)

        if ofmt == "text":
            formatting["cmd_type"] = "cli_show_ascii"
            formatting["ofmt"] = "xml"

        nxapi_cmd = self.nxapi_command(
            " ;".join([command] if command else commands), formatting
        )
        res = await self.nxapi_exec(nxapi_cmd, formatting=formatting, strip_ns=strip_ns)
        return res[0] if command else res

    async def check_connection(self) -> bool:
        """
        This function checks the target device to ensure that the NXAPI port is
        open and accepting connections.  It is recommended that a Caller checks
        the connection before involing cli commands, but this step is not
        required.

        Returns
        -------
        True when the device NXAPI is accessible, False otherwise.
        """
        try:
            await asyncio.open_connection(self.host, port=self.port)
        except OSError:
            return False
        return True

    # async def nxapi_write_config(self, xcmd):
    #     """
    #     This coroutine is used to push the configuration to the device and
    #     return any error XML elements.  If no errors then return value is None.
    #     """
    #     res = await self.post("/ins", data=xcmd)
    #     res.raise_for_status()
    #     as_xml = xmlhelp.fromstring(res.text)
    #
    #     if any_errs := as_xml.xpath(".//code[. != '200']"):
    #         return any_errs
    #
    #     return None



from lxml import etree

_xparser = etree.XMLParser(recover=True)


def strip_ns(root: etree.Element) -> etree.Element:
    """
    This function removes all namespace information from an XML Element tree
    so that a Caller can then use the `xpath` function without having
    to deal with the complexities of namespaces.
    """

    # first we visit each node in the tree and set the tag name to its localname
    # value; thus removing its namespace prefix

    for elem in root.getiterator():
        elem.tag = etree.QName(elem).localname

    # at this point there are no tags with namespaces, so we run the cleanup
    # process to remove the namespace definitions from within the tree.

    etree.cleanup_namespaces(root)
    return root


def fromstring(content):
    return etree.fromstring(content, parser=_xparser)


def tostring(ele):
    return etree.tostring(ele, pretty_print=True).decode()



"""
This module provides the asyncio client for Cisco NX-API, expected to work on:
    * N3K system
    * N5K system  (tested)
    * N7K system
    * N9K system
"""

from typing import Optional, AnyStr, Tuple, List

import json
from socket import getservbyname
import httpx
import base64
import ssl
from collections import namedtuple

from . import xmlhelp


__all__ = ["Device", "CommandResults"]

_ssl_context = ssl.create_default_context()
_ssl_context.options &= ~ssl.OP_NO_TLSv1_2 & ~ssl.OP_NO_TLSv1_1
_ssl_context.minimum_version = ssl.TLSVersion.TLSv1
_ssl_context.check_hostname = False
_ssl_context.verify_mode = ssl.CERT_NONE
_ssl_context.set_ciphers("HIGH:!DH:!aNULL")

# _ssl_context = ssl.SSLContext(ssl_version=ssl.PROTOCOL_TLSv1_1)  # noqa

# _ssl_context = ssl.create_default_context()
# # Sets up old and insecure TLSv1.
# _ssl_context.options &= ~ssl.OP_NO_TLSv1_3 & ~ssl.OP_NO_TLSv1_2 & ~ssl.OP_NO_TLSv1_1
# _ssl_context.minimum_version = ssl.TLSVersion.TLSv1


_NXAPI_CMD_TEMPLATE = """\
<?xml version="1.0"?>
<ins_api>
<version>{api_ver}</version>
<type>{cmd_type}</type>
<chunk>{chunk}</chunk>
<sid>{sid}</sid>
<input>{cmd_input}</input>
<output_format>{ofmt}</output_format>
</ins_api>"""

CommandResults = namedtuple("CommandResults", ["ok", "command", "output"])


class Transport(object):
    CMDTYPES_OPTIONS = ("cli_show", "cli_show_ascii", "cli_conf", "bash")
    OFMT_OPTIONS = ("xml", "json", "text")
    API_VER = "1.0"

    def __init__(self, host, proto, port, creds, timeout=60):
        port = port or getservbyname(proto)

        self.client = httpx.AsyncClient(
            base_url=httpx.URL(f"{proto}://{host}:{port}"),
            verify=_ssl_context,
            timeout=httpx.Timeout(timeout),
        )

        self.client.headers["Content-Type"] = "application/xml"
        self.b64auth = (
            base64.encodebytes(bytes("%s:%s" % creds, encoding="utf-8"))
            .decode()
            .replace("\n", "")
        )
        self.client.headers["Authorization"] = "Basic %s" % self.b64auth

        self.username = creds[0]
        self.cmd_type = "cli_show"
        self.ofmt = "xml"

    @property
    def timeout(self):
        return self.client.timeout

    @timeout.setter
    def timeout(self, value):
        self.client.timeout = httpx.Timeout(value)

    def form_command(self, cmd_input, formatting, sid=None):
        cmd_type = formatting.setdefault("cmd_type", self.cmd_type)
        ofmt = formatting.setdefault("ofmt", self.ofmt)

        return _NXAPI_CMD_TEMPLATE.format(
            api_ver=self.API_VER,
            cmd_type=cmd_type or self.cmd_type,
            cmd_input=cmd_input,
            chunk=0,
            sid=sid or "sid",
            ofmt=ofmt or self.ofmt,
        )

    async def post(self, xcmd, formatting, strip_ns=False):
        res = await self.client.post("/ins", data=xcmd)
        res.raise_for_status()

        if formatting["ofmt"] == "json":
            as_json = json.loads(res.text)

            outputs = as_json["ins_api"]["outputs"]["output"]
            if not isinstance(outputs, list):
                outputs = [outputs]

            return [
                CommandResults(
                    ok=cmd_res["code"] == "200",
                    command=cmd_res["input"],
                    output=cmd_res["body"],
                )
                for cmd_res in outputs
            ]

        # Output format is "xml" or "text"; but in either case the body content
        # is extracted in the same manner.

        as_text = xmlhelp.strip_ns(res.text) if strip_ns else res.text
        as_xml = xmlhelp.fromstring(as_text)

        def body_is_text(_res_e):
            return _res_e.find("body").text.strip()

        def body_is_xml(_res):
            return _res.find("body")

        get_output = (
            body_is_text if formatting["cmd_type"] == "cli_show_ascii" else body_is_xml
        )

        return [
            CommandResults(
                ok=cmd_res.findtext("code") == "200",
                command=cmd_res.findtext("input").strip(),
                output=get_output(cmd_res),
            )
            for cmd_res in as_xml.xpath("outputs/output")
        ]

    async def post_write_config(self, xcmd):
        """
        This coroutine is used to push the configuration to the device an return any
        error XML elements.  If no errors then return value is None.
        """
        res = await self.client.post("/ins", data=xcmd)
        res.raise_for_status()
        as_xml = xmlhelp.fromstring(res.text)

        if any_errs := as_xml.xpath(".//code[. != '200']"):
            return any_errs

        return None


class Device(object):
    def __init__(
        self,
        host: AnyStr,
        creds: Tuple[str, str],
        proto: Optional[AnyStr] = "https",
        port=None,
        private=None,
    ):
        self.api = Transport(host=host, creds=creds, proto=proto, port=port)
        self.private = private
        self.host = host

    async def exec(
        self, commands: List[AnyStr], ofmt: Optional[AnyStr] = None, strip_ns=False
    ) -> List[CommandResults]:
        """
        Execute a list of operational commands and return the output as a list of CommandResults.
        """
        formatting = dict(ofmt=ofmt)

        if ofmt == "text":
            formatting["cmd_type"] = "cli_show_ascii"
            formatting["ofmt"] = "xml"

        xcmd = self.api.form_command(" ;".join(commands), formatting)
        return await self.api.post(xcmd, formatting, strip_ns=strip_ns)

    async def push_config(self, content: AnyStr):
        xcmd = self.api.form_command(
            cmd_input=" ; ".join(content.splitlines()),
            formatting=dict(cmd_type="cli_conf", ofmt="xml"),
        )
        return await self.api.post_write_config(xcmd)

    async def get_config(self, ofmt="text"):
        res = await self.exec(["show running-config"], ofmt=ofmt, strip_ns=True)
        return res[0].output



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aionxapi', 'asyncnxapi']

package_data = \
{'': ['*']}

install_requires = \
['httpx>=0.23.0,<0.24.0', 'lxml>=4.9.1,<5.0.0']

setup_kwargs = {
    'name': 'aio-nxapi',
    'version': '0.5.0',
    'description': 'Cisco NX-API asyncio client',
    'long_description': '# Cisco NX-API asyncio Client\n\nThis repository contains a Cisco NX-API asyncio based client that uses\nthe [httpx](https://www.python-httpx.org/) as an underlying transport and\n[lxml](https://lxml.de/) as the basis for handling XML.\n\nNote: This client does not support the NETCONF interface.\n\n**WORK IN PROGESS**\n\n### Quick Example\n\nThie following shows how to create a Device instance and run a list of\ncommands.\n\nBy default the Device instance will use HTTPS transport.  The Device instance\nsupports the following settings:\n\n   * `host` - The device hostname or IP address\n   * `username` - The login user-name\n   * `password` - The login password\n   * `proto` - *(Optional)* Choose either "https" or "http", defaults to "https"\n   * `port` - *(Optional)* Chose the protocol port to override proto default\n\nThe result of command execution is a list of CommandResults (namedtuple).\nThe `output` field will be:\n   * lxml.Element when output format is \'xml\'\n   * dict when output format is \'json\'\n   * str when output format is \'text\'\n\n```python\nfrom asyncnxapi import Device\n\nusername = \'dummy-user\'\npassword = \'dummy-password\'\n\nasync def run_test(host):\n    dev = Device(host=host, creds=(username, password))\n    res = await dev.exec([\'show hostname\', \'show version\'], ofmt=\'json\')\n    for cmd in res:\n       if not cmd.ok:\n          print(f"{cmd.command} failed")\n          continue\n\n       # do something with cmd.output as dict since ofmt was \'json\'\n```\n\n# Limitations\n\n  * Chunking is not currently supported.  If anyone has need of this feature\n  please open an issue requesting support.\n\n# References\n\nCisco DevNet NX-API Rerefence:<br/>\n   * https://developer.cisco.com/site/cisco-nexus-nx-api-references/\n\nCisco platform specific NX-API references:\n\n   * N3K systems, requires 7.0(3)I2(2) or later:\n    https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus3000/sw/programmability/7_x/b_Cisco_Nexus_3000_Series_NX-OS_Programmability_Guide_7x/b_Cisco_Nexus_3000_Series_NX-OS_Programmability_Guide_7x_chapter_010010.html\n\n   * N5K system, requires 7.3(0)N1(1) or later:\n    https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus5000/sw/programmability/guide/b_Cisco_Nexus_5K6K_Series_NX-OS_Programmability_Guide/nx_api.html#topic_D110A801F14F43F385A90DE14293BA46\n\n   * N7K systems:\n    https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus7000/sw/programmability/guide/b_Cisco_Nexus_7000_Series_NX-OS_Programmability_Guide/b_Cisco_Nexus_7000_Series_NX-OS_Programmability_Guide_chapter_0101.html\n\n   * N9K systems:\n    https://www.cisco.com/c/en/us/td/docs/switches/datacenter/nexus9000/sw/6-x/programmability/guide/b_Cisco_Nexus_9000_Series_NX-OS_Programmability_Guide/b_Cisco_Nexus_9000_Series_NX-OS_Programmability_Guide_chapter_011.html\n',
    'author': 'Jeremy Schulman',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
